"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.useLlmInference = useLlmInference;
var _react = _interopRequireDefault(require("react"));
var _reactNative = require("react-native");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const {
  LlmInferenceModule
} = _reactNative.NativeModules;
const eventEmitter = new _reactNative.NativeEventEmitter(LlmInferenceModule);
function getLlmInference() {
  return LlmInferenceModule;
}
eventEmitter.addListener('logging', ev => {
  console.log(`[${ev.handle}] ${ev.message}`);
});
function getConfigStorageKey(config) {
  if (config.storageType === 'asset') {
    return `${config.modelName}`;
  }
  return `${config.modelPath}`;
}
function useLlmInference(config) {
  const [modelHandle, setModelHandle] = _react.default.useState();
  const configStorageKey = getConfigStorageKey(config);
  _react.default.useEffect(() => {
    // Skip model creation if configStorageKey is empty
    if (configStorageKey.length === 0) {
      console.warn('No valid model path or name provided. Skipping model creation.');
      return;
    }
    let newHandle;
    const modelCreatePromise = config.storageType === 'asset' ? getLlmInference().createModelFromAsset(configStorageKey, config.maxTokens ?? 512, config.topK ?? 40, config.temperature ?? 0.8, config.randomSeed ?? 0) : getLlmInference().createModel(configStorageKey, config.maxTokens ?? 512, config.topK ?? 40, config.temperature ?? 0.8, config.randomSeed ?? 0);
    modelCreatePromise.then(handle => {
      console.log(`Created model with handle ${handle}`);
      setModelHandle(handle);
      newHandle = handle;
    }).catch(error => {
      console.error('createModel', error);
    });
    return () => {
      if (newHandle !== undefined) {
        getLlmInference().releaseModel(newHandle).then(() => {
          console.log(`Released model with handle ${newHandle}`);
        }).catch(() => {
          console.error('releaseModel');
        });
      }
    };
  }, [config.maxTokens, config.storageType, config.randomSeed, config.temperature, config.topK, configStorageKey]);
  const nextRequestIdRef = _react.default.useRef(0);
  const generateResponse = _react.default.useCallback(async (prompt, onPartial, onError, abortSignal) => {
    if (modelHandle === undefined) {
      throw new Error('Model handle is not defined');
    }
    const requestId = nextRequestIdRef.current++;
    const partialSub = eventEmitter.addListener('onPartialResponse', ev => {
      // console.log(
      //   `[${ev.handle}] partial response ${ev.requestId}: ${ev.partial}`
      // );
      console.log(JSON.stringify(ev));
      if (onPartial && requestId === ev.requestId && !(abortSignal?.aborted ?? false)) {
        onPartial(ev.response, ev.requestId);
      }
    });
    const errorSub = eventEmitter.addListener('onErrorResponse', ev => {
      console.log(`[${ev.handle}] error ${ev.requestId}: ${ev.error}`);
      if (onError && requestId === ev.requestId && !(abortSignal?.aborted ?? false)) {
        onError(ev.error, ev.requestId);
      }
    });
    try {
      return await getLlmInference().generateResponse(modelHandle, requestId, prompt);
    } catch (e) {
      console.error(e);
      throw e;
    } finally {
      console.log('finally: removing listeners');
      partialSub.remove();
      errorSub.remove();
    }
  }, [modelHandle]);
  return _react.default.useMemo(() => ({
    generateResponse,
    isLoaded: modelHandle !== undefined
  }), [generateResponse, modelHandle]);
}

// class InferenceError extends Error {
//   code: string;

//   constructor(message: string, code: string) {
//     super(message); // Pass the message to the Error constructor
//     this.code = code; // Initialize the custom code property
//     this.name = "InferenceError"; // Set the error name as the class name

//     // Maintaining proper stack trace for where the error was thrown (only available on V8)
//     if (Error.captureStackTrace) {
//       Error.captureStackTrace(this, InferenceError);
//     }
//   }
// }
//# sourceMappingURL=llmInference.js.map